/**
 * @file component.c
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "utils.h"
#include "yac_mpi.h"
#include "component.h"
#include "yac_interface.h"

// general component struct
struct component {

  // general component data

  char * name;
  unsigned int global_comp_id;
  MPI_Comm comm;
  unsigned io_proc_step;
};

struct component_groups {
  char const * name;
  MPI_Group group;
};

struct yac_component_config {

  MPI_Comm comm;

  struct component ** components;
  size_t num_components;

  struct component_groups * comp_groups;
  size_t num_comp_groups;
};


static unsigned compute_io_proc_step(unsigned comm_size) {

  unsigned num_procs_per_node = (unsigned)-1;
  // last entry has to be 1
  unsigned potential_node_sizes[] = {36, 24, 16, 12, 8, 6, 4, 2, 1};

  // estimate the number of processes per node
  for (size_t i = 0;
       (i < sizeof(potential_node_sizes) / sizeof(potential_node_sizes[0])) &&
       (num_procs_per_node == (unsigned)-1); ++i)
    if (comm_size % potential_node_sizes[i] == 0)
      num_procs_per_node = potential_node_sizes[i];

  YAC_ASSERT(num_procs_per_node != (unsigned)-1, "internal error")

  unsigned num_nodes = comm_size / num_procs_per_node;
  unsigned num_io_procs = MIN((unsigned)sqrt((float)num_nodes), 64);
  return ((num_nodes + num_io_procs - 1) / num_io_procs) * num_procs_per_node;
}

struct yac_component_config * yac_component_config_new(
  struct yac_couple_config * couple_config, char const ** names,
  size_t num_names, MPI_Comm comm) {

  for (size_t i = 0; i < num_names; ++i)
    YAC_ASSERT_F(
      yac_couple_config_component_name_is_valid(couple_config, names[i]),
      "ERROR(yac_component_config_new): "
      "component \"%s\" not in coupling configuration", names[i])

  struct yac_component_config * comp_config = xmalloc(1 * sizeof(*comp_config));

  yac_mpi_call(MPI_Comm_dup(comm, &(comp_config->comm)), comm);
  comm = comp_config->comm;
  comp_config->components =
    xcalloc(num_names, sizeof(*(comp_config->components)));
  comp_config->num_components = num_names;

  size_t num_comp_groups =
    ((comp_config->num_comp_groups =
        yac_couple_config_get_num_components(couple_config)));

  unsigned * global_comp_ids = xmalloc(num_names * sizeof(*global_comp_ids));
  int * idx = xmalloc(num_names * sizeof(*idx));

  // for all names get component ids from coupling configuration
  for (unsigned i = 0; i < num_names; ++i) {
    global_comp_ids[i] =
      yac_couple_config_get_component_id_by_name(couple_config, names[i]);
    YAC_ASSERT_F(
      global_comp_ids[i] != (unsigned)-1,
      "ERROR(yac_component_config_new): mismatch in component name\n"
      "Check def_comp call and coupling configuration for %s!", names[i])
    idx[i] = i;
  }

  yac_quicksort_index((int*)global_comp_ids, (size_t)num_names, idx);

  // check for duplicated component ids
  for (unsigned i = 1; i < num_names; ++i)
    YAC_ASSERT(
      global_comp_ids[i-1] != global_comp_ids[i],
      "ERROR(yac_component_config_new): duplicated component ids")

  struct component_groups * comp_groups =
    ((comp_config->comp_groups =
        xmalloc(num_comp_groups * sizeof(*(comp_config->comp_groups)))));

  MPI_Group world_group;
  yac_mpi_call(MPI_Comm_group(comm, &world_group), comm);

  // for all components in coupling configuration
  for (size_t i = 0, comp_idx = 0; i < num_comp_groups; ++i) {
    int curr_comp_id = yac_couple_config_get_component_id(couple_config, i);
    comp_groups[i].name =
      yac_couple_config_get_component_name(couple_config, i);

    unsigned curr_comp_idx;
    for (curr_comp_idx = 0; curr_comp_idx < num_names; ++curr_comp_idx)
      if (global_comp_ids[curr_comp_idx] == curr_comp_id) break;

    // if the local process has the current component
    if (curr_comp_idx < num_names) {

      struct component * comp = xmalloc(1 * sizeof(*comp));

      comp_config->components[comp_idx++] = comp;

      char const * curr_comp_name = names[idx[curr_comp_idx]];
      comp->name = xmalloc(strlen(curr_comp_name) + 1);
      strcpy(comp->name, curr_comp_name);

      comp->global_comp_id = curr_comp_id;
      yac_mpi_call(MPI_Comm_split(comm, 0, 0, &(comp->comm)), comm);

      // create group for all processes in the component communicator
      MPI_Group comp_group;
      yac_mpi_call(MPI_Comm_group(comp->comm, &comp_group), comp->comm);
      int comp_group_size;
      yac_mpi_call(MPI_Group_size(comp_group, &comp_group_size), comp->comm);
      int * rank_buffer =
        xmalloc(2 * (size_t)comp_group_size * sizeof(*rank_buffer));
      int * comp_ranks = rank_buffer,
          * world_ranks = rank_buffer + (size_t)comp_group_size;
      for (int j = 0; j < comp_group_size; ++j) comp_ranks[j] = j;
      yac_mpi_call(
        MPI_Group_translate_ranks(
          comp_group, comp_group_size, comp_ranks,
          world_group, world_ranks), comp->comm);
      yac_mpi_call(
        MPI_Group_incl(world_group, comp_group_size, world_ranks,
                       &(comp_groups[i].group)), comp->comm);
      yac_mpi_call(MPI_Group_free(&comp_group), comm);
      free(rank_buffer);

      comp->io_proc_step = compute_io_proc_step(comp_group_size);

    } else {

      MPI_Comm dummy_comm;
      yac_mpi_call(MPI_Comm_split(comm, 1, 0, &dummy_comm), comm);

      // create group for all processes in the component communicator
      MPI_Group non_comp_group;
      yac_mpi_call(MPI_Comm_group(dummy_comm, &non_comp_group), dummy_comm);
      int non_comp_group_size;
      yac_mpi_call(
        MPI_Group_size(non_comp_group, &non_comp_group_size), dummy_comm);
      int * rank_buffer =
        xmalloc(2 * (size_t)non_comp_group_size * sizeof(*rank_buffer));
      int * non_comp_ranks = rank_buffer,
          * world_ranks = rank_buffer + (size_t)non_comp_group_size;
      for (int j = 0; j < non_comp_group_size; ++j) non_comp_ranks[j] = j;
      yac_mpi_call(
        MPI_Group_translate_ranks(
          non_comp_group, non_comp_group_size, non_comp_ranks,
          world_group, world_ranks), dummy_comm);
      yac_mpi_call(
        MPI_Group_excl(world_group, non_comp_group_size, world_ranks,
                       &(comp_groups[i].group)), comm);
      yac_mpi_call(MPI_Group_free(&non_comp_group), comm);
      free(rank_buffer);
      yac_mpi_call(MPI_Comm_free(&dummy_comm), comm);
    }
  }

  yac_mpi_call(MPI_Group_free(&world_group), comm);

  free(idx);
  free(global_comp_ids);

  return comp_config;
}

const char * yac_component_get_name(struct component * comp) {

  YAC_ASSERT(comp != NULL, "ERROR: invalid component pointer")

  return comp->name;
}

unsigned int yac_component_get_global_id(struct component * comp) {

  YAC_ASSERT(comp != NULL, "ERROR: invalid component pointer")

   return comp->global_comp_id;
}

MPI_Comm yac_component_get_communicator(struct component * comp) {

  YAC_ASSERT(comp != NULL, "ERROR: invalid component pointer")

   return comp->comm;
}

void yac_component_set_io_proc_step(struct component * comp,
                                    unsigned io_proc_step) {

  YAC_ASSERT(comp != NULL, "ERROR: invalid component pointer")
  YAC_ASSERT(io_proc_step != 0, "ERROR: invalid io_proc_step")

   comp->io_proc_step = io_proc_step;
}

unsigned yac_component_get_io_proc_step(struct component * comp) {

  YAC_ASSERT(comp != NULL, "ERROR: invalid component pointer")

   return comp->io_proc_step;
}

MPI_Comm yac_component_config_get_pair_rootcomm(
  struct yac_component_config * comp_config,
  const char * comp_name1, const char * comp_name2) {

  YAC_ASSERT(
    strcmp(comp_name1, comp_name2),
    "ERROR(yac_component_config_get_pair_rootcomm): "
    "both component names are identical")

  size_t comp_idx1 = SIZE_MAX, comp_idx2 = SIZE_MAX;

  for (size_t i = 0; i < comp_config->num_comp_groups; ++i) {

    if (!strcmp(comp_config->comp_groups[i].name, comp_name1)) comp_idx1 = i;
    if (!strcmp(comp_config->comp_groups[i].name, comp_name2)) comp_idx2 = i;
  }

  if (comp_idx1 > comp_idx2) {
    size_t temp = comp_idx1;
    comp_idx1 = comp_idx2;
    comp_idx2 = temp;
  }

  YAC_ASSERT_F(
    (comp_idx1 != SIZE_MAX) && (comp_idx2 != SIZE_MAX),
    "ERROR(yac_component_config_get_pair_rootcomm): "
    "invalid component name: \"%s\"",
    (comp_idx1 == SIZE_MAX)?comp_name1:comp_name2);

  YAC_ASSERT(
    (comp_idx1 <= INT_MAX) && (comp_idx2 <= INT_MAX),
    "ERROR(yac_component_config_get_pair_rootcomm): invalid component index");

  MPI_Group root_group;

  yac_mpi_call(
    MPI_Group_union(
      comp_config->comp_groups[comp_idx1].group,
      comp_config->comp_groups[comp_idx2].group,
      &root_group), comp_config->comm);

  int group_rank, group_size;
  yac_mpi_call(MPI_Group_rank(root_group, &group_rank), comp_config->comm);
  yac_mpi_call(MPI_Group_size(root_group, &group_size), comp_config->comm);
   YAC_ASSERT(
    group_rank != MPI_UNDEFINED,
    "ERROR(yac_component_config_get_pair_rootcomm): "
    "local process not included in any component provided to this routine");

  // get rank (from comp_config->comm) of neighbouring ranks
  int group_neigh_ranks[2], neigh_ranks[2];
  group_neigh_ranks[0] = (group_rank + 1) % group_size;
  group_neigh_ranks[1] = (group_rank + group_size - 1) % group_size;
  MPI_Group comp_config_group;
  yac_mpi_call(
    MPI_Comm_group(comp_config->comm, &comp_config_group), comp_config->comm);
  yac_mpi_call(
    MPI_Group_translate_ranks(
      root_group, 2, group_neigh_ranks, comp_config_group, neigh_ranks),
    comp_config->comm);
  MPI_Group_free(&comp_config_group);

  // exchange component indices with neighbouring processes
  int const tag = 0;
  int comp_idx_buffer[2];
  comp_idx_buffer[0] = comp_idx1;
  comp_idx_buffer[1] = comp_idx2;
  yac_mpi_call(
    MPI_Sendrecv_replace(
      comp_idx_buffer, 2, MPI_INT, neigh_ranks[0], tag, neigh_ranks[1], tag,
      comp_config->comm, MPI_STATUS_IGNORE), comp_config->comm);
  YAC_ASSERT_F(
    (comp_idx_buffer[0] == (int)comp_idx1) &&
    (comp_idx_buffer[1] == (int)comp_idx2),
    "ERROR(yac_component_config_get_pair_rootcomm): "
    "processes do not agree on components "
    "(local component names: \"%s\" and \"%s\")", comp_name1, comp_name2);

  MPI_Comm root_comm;
  yac_mpi_call(
    MPI_Comm_create_group(
      comp_config->comm, root_group, tag, &root_comm), comp_config->comm);

  yac_mpi_call(MPI_Group_free(&root_group), comp_config->comm);

  return root_comm;
}

static void yac_component_delete(struct component * comp) {

  YAC_ASSERT(comp != NULL, "ERROR: invalid component pointer")

  free(comp->name);
  yac_mpi_call(MPI_Comm_free(&(comp->comm)), MPI_COMM_WORLD);
  free(comp);
}

struct component ** yac_component_config_get_components(
  struct yac_component_config * comp_config) {

  return comp_config->components;
}

size_t yac_component_config_get_num_components(
  struct yac_component_config * comp_config) {

  return comp_config->num_components;
}

void yac_component_config_delete(struct yac_component_config * comp_config) {

  if (comp_config == NULL) return;

  for (size_t i = 0; i < comp_config->num_components; ++i)
    yac_component_delete(comp_config->components[i]);
  free(comp_config->components);

  for (size_t i = 0; i < comp_config->num_comp_groups; ++i)
    yac_mpi_call(
      MPI_Group_free(&(comp_config->comp_groups[i].group)), comp_config->comm);
  free(comp_config->comp_groups);

  yac_mpi_call(MPI_Comm_free(&(comp_config->comm)), MPI_COMM_WORLD);
  free(comp_config);
}
