/**
 * @file yac_interface.h
 * @brief C User Interface
 *
 * An overview of the C User Interface can be found \ref cinterface "here".
 * For a detailed description of the Fortran interface please see
 * \ref mo_yac_finterface.f90 and \ref yac_finterface.f90
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *                                 Teresa Holfeld <teresa.holfeld@zmaw.de>
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 *         Teresa Holfeld <teresa.holfeld@zmaw.de>
 * @par Warranty
 *
 * This code has been tested up to a certain level. Defects and weaknesses,
 * which may be included in the code, do not establish any warranties by the
 * authors. The authors do not make any warranty, express or implied, or assume any
 * liability or responsibility for the use, acquisition or application of this
 * software.
 *
 * @par C User API
 *
 * For a detailed description of the C interface please see \ref cinterface.
 * Information on the Fortran interface can be found here \ref finterface.
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 *             Teresa Holfeld <teresa.holfeld@zmaw.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef YAC_INTERFACE_H
#define YAC_INTERFACE_H

#include <mpi.h>

extern int const YAC_LOCATION_CELL;
extern int const YAC_LOCATION_CORNER;
extern int const YAC_LOCATION_EDGE;

extern int const YAC_EXCHANGE_TYPE_NONE;
extern int const YAC_EXCHANGE_TYPE_SOURCE;
extern int const YAC_EXCHANGE_TYPE_TARGET;

extern int const YAC_ACTION_NONE;
extern int const YAC_ACTION_REDUCTION;
extern int const YAC_ACTION_COUPLING;
extern int const YAC_ACTION_GET_FOR_RESTART;
extern int const YAC_ACTION_PUT_FOR_RESTART;
extern int const YAC_ACTION_OUT_OF_BOUND;

extern int const YAC_REDUCTION_INSTANT;
extern int const YAC_REDUCTION_ACCUMULATE;
extern int const YAC_REDUCTION_TIME_AVERAGE;
extern int const YAC_REDUCTION_TIME_MINIMUM;
extern int const YAC_REDUCTION_TIME_MAXIMUM;

extern int const YAC_TIME_UNIT_MILLISECOND;
extern int const YAC_TIME_UNIT_SECOND;
extern int const YAC_TIME_UNIT_MINUTE;
extern int const YAC_TIME_UNIT_HOUR;
extern int const YAC_TIME_UNIT_DAY;
extern int const YAC_TIME_UNIT_MONTH;
extern int const YAC_TIME_UNIT_YEAR;
extern int const YAC_TIME_UNIT_ISO_FORMAT;

extern int const YAC_CALENDAR_NOT_SET;
extern int const YAC_PROLEPTIC_GREGORIAN;
extern int const YAC_YEAR_OF_365_DAYS;
extern int const YAC_YEAR_OF_360_DAYS;

extern int const YAC_AVG;
extern int const YAC_DIST;
extern int const YAC_BARY;

extern int const YAC_NNN_AVG;
extern int const YAC_NNN_DIST;
extern int const YAC_NNN_GAUSS;
extern int const YAC_NNN_RBF;

extern int const YAC_CONSERV_DESTAREA;
extern int const YAC_CONSERV_FRACAREA;

extern int const YAC_SPMAP_AVG;
extern int const YAC_SPMAP_DIST;

extern int const COUPLE_CONFIG_DUMMY_ID;

#define YAC_MAX_CHARLEN (132)

/** \example test_dummy_coupling_c.c
 * This example simulates a whole model setup with three components (ocean,
 * atmosphere, io). It uses one process for each component.
 */

/** \example test_dummy_coupling2_c.c
 * This example simulates a whole model setup with two components (ocean,
 * atmosphere) where rank 2 is shared by both components.
 */

/** \example test_dummy_coupling3_c.c
 * This tests checks the support for multiple parallel YAC instances.
 * There are three main components and one dummy component. The dummy component
 * does not take part in any coupling. YAC instances are used.
 */

/** \example test_dummy_coupling4_c.c
 */

/** \example test_dummy_coupling5_c.c
 */

/** \example test_dummy_coupling6_c.c
 */

/** \example test_restart_c.c
 * This example simulates a restart of a
 * coupled configuration with a grid redefiniton.
 */

/** \example test_restart2.c
 * This example simulates a bit more complex restart
 * of a coupled configuration with a grid redefiniton.
 */

/** \example test_init_group_c.c
 */

/* -----------------------------------------------------------------------------

    C and Fortran description of the user API

   ----------------------------------------------------------------------------- */

/** Elementary initialisation of the whole system

     @param[in] xml_filename    coupling xml configuration
     @param[in] schema_filename corresponding ml schema file
     @remark this call initialises the default YAC instance
     @remark this call is collective for all processes in MPI_COMM_WORLD
     @remark In case a process in MPI_COMM_WORLD does not wish to initialise
             YAC, it can call \ref yac_cinit_dummy instead.
     @remark At least one process has to provide valid arguments for
             xml_filename and schema_filename. All other ranks can provide
             NULL instead. If multiple ranks provide valid file names, the one
             with the lowest rank in MPI_COMM_WORLD will read and distribute
             the coupling configuration.
*/
    void yac_cinit ( const char * xml_filename,
                     const char * schema_filename );

/** Elementary initialisation of the whole system

     @param[in] xml_filename    coupling xml configuration
     @param[in] schema_filename corresponding ml schema file
     @param[in] group_name      name of a group of processes
                                (NULL is a valid argument)
     @remark this call initialises the default YAC instance
     @remark this call is collective for all processes in MPI_COMM_WORLD
     @remark In case a process in MPI_COMM_WORLD does not wish to initialise
             YAC, it can call \ref yac_cinit_dummy instead.
     @remark YAC will generate a communicator for all processes in comm that
             provid the same non-NULL group name. This communicator can be
             retrieved by \ref yac_cget_groupcomm.
     @remark At least one process has to provide valid arguments for
             xml_filename and schema_filename. All other ranks can provide
             NULL instead. If multiple ranks provide valid file names, the one
             with the lowest rank in MPI_COMM_WORLD will read and distribute
             the coupling configuration.
*/
    void yac_cinit_group ( const char * xml_filename,
                           const char * schema_filename,
                           const char * group_name );

/** Elementary initialisation of the whole system

     @param[in]  xml_filename    coupling xml configuration
     @param[in]  schema_filename corresponding ml schema file
     @param[out] yac_instance_id id of the YAC instance initialised by
                                 this call
     @remark this call is collective for all processes in MPI_COMM_WORLD
     @remark In case a process in MPI_COMM_WORLD does not wish to initialise
             YAC, it can call \ref yac_cinit_dummy instead.
     @remark At least one process has to provide valid arguments for
             xml_filename and schema_filename. All other ranks can provide
             NULL instead. If multiple ranks provide valid file names, the one
             with the lowest rank in MPI_COMM_WORLD will read and distribute
             the coupling configuration.
*/
    void yac_cinit_instance ( const char * xml_filename,
                              const char * schema_filename,
                              int * yac_instance_id );

/** Elementary initialisation of the whole system

     @param[in]  xml_filename    coupling xml configuration
     @param[in]  schema_filename corresponding ml schema file
     @param[in]  group_name      name of a group of processes
                                 (NULL is a valid argument)
     @param[out] yac_instance_id id of the YAC instance initialised by
                                 this call
     @remark this call is collective for all processes in MPI_COMM_WORLD
     @remark In case a process in MPI_COMM_WORLD does not wish to initialise
             YAC, it can call \ref yac_cinit_dummy instead.
     @remark YAC will generate a communicator for all processes in comm that
             provid the same non-NULL group name. This communicator can be
             retrieved by \ref yac_cget_groupcomm_instance.
     @remark At least one process has to provide valid arguments for
             xml_filename and schema_filename. All other ranks can provide
             NULL instead. If multiple ranks provide valid file names, the one
             with the lowest rank in MPI_COMM_WORLD will read and distribute
             the coupling configuration.
*/
    void yac_cinit_group_instance ( const char * xml_filename,
                                    const char * schema_filename,
                                    const char * group_name,
                                    int * yac_instance_id );

/** Elementary initialisation of the whole system using a user-provided
 *  MPI communicator

     @param[in] xml_filename    coupling xml configuration
     @param[in] schema_filename corresponding ml schema file
     @param[in] comm            MPI communicator
     @remark this call initialises the default YAC instance
     @remark the MPI communicator provided to this routine has to contain
             all processes that will take part in the coupling
     @remark this call is collective for all processes in comm
     @remark In case a process in comm does not wish to initialise
             YAC, it can call \ref yac_cinit_comm_dummy instead.
     @remark At least one process has to provide valid arguments for
             xml_filename and schema_filename. All other ranks can provide
             NULL instead. If multiple ranks provide valid file names, the one
             with the lowest rank in MPI_COMM_WORLD will read and distribute
             the coupling configuration.
*/
    void yac_cinit_comm ( const char * xml_filename,
                          const char * schema_filename,
                          MPI_Comm comm );

/** Elementary initialisation of the whole system using a user-provided
 *  MPI communicator

     @param[in] xml_filename    coupling xml configuration
     @param[in] schema_filename corresponding ml schema file
     @param[in] comm            MPI communicator
     @param[in] group_name      name of a group of processes
                                (NULL is a valid argument)
     @remark this call initialises the default YAC instance
     @remark the MPI communicator provided to this routine has to contain
             all processes that will take part in the coupling
     @remark this call is collective for all processes in comm
     @remark In case a process in comm does not wish to initialise
             YAC, it can call \ref yac_cinit_comm_dummy instead.
     @remark YAC will generate a communicator for all processes in comm that
             provid the same non-NULL group name. This communicator can be
             retrieved by \ref yac_cget_groupcomm.
     @remark At least one process has to provide valid arguments for
             xml_filename and schema_filename. All other ranks can provide
             NULL instead. If multiple ranks provide valid file names, the one
             with the lowest rank in comm will read and distribute
             the coupling configuration.
*/
    void yac_cinit_comm_group ( const char * xml_filename,
                                const char * schema_filename,
                                MPI_Comm comm,
                                const char * group_name );

/** Elementary initialisation of the whole system using a user-provided
 *  MPI communicator

     @param[in]  xml_filename    coupling xml configuration
     @param[in]  schema_filename corresponding ml schema file
     @param[in]  comm            MPI communicator
     @param[out] yac_instance_id id of the YAC instance initialised by
                                 this call
     @remark In case the user has multiple YAC instances in parallel, he has
             to initialise YAXT himself.
     @remark In case a process in comm does not wish to initialise
             YAC, it can call \ref yac_cinit_comm_dummy instead.
     @remark At least one process has to provide valid arguments for
             xml_filename and schema_filename. All other ranks can provide
             NULL instead. If multiple ranks provide valid file names, the one
             with the lowest rank in comm will read and distribute
             the coupling configuration.
*/
    void yac_cinit_comm_instance ( const char * xml_filename,
                                   const char * schema_filename,
                                   MPI_Comm comm, int * yac_instance_id );

/** Elementary initialisation of the whole system using a user-provided
 *  MPI communicator

     @param[in]  xml_filename    coupling xml configuration
     @param[in]  schema_filename corresponding ml schema file
     @param[in]  comm            MPI communicator
     @param[in]  group_name      name of a group of processes
                                 (NULL is a valid argument)
     @param[out] yac_instance_id id of the YAC instance initialised by
                                 this call
     @remark In case the user has multiple YAC instances in parallel, he has
             to initialise YAXT himself.
     @remark In case a process in comm does not wish to initialise
             YAC, it can call \ref yac_cinit_comm_dummy instead.
     @remark YAC will generate a communicator for all processes in comm that
             provid the same non-NULL group name. This communicator can be
             retrieved by \ref yac_cget_groupcomm_instance.
     @remark At least one process has to provide valid arguments for
             xml_filename and schema_filename. All other ranks can provide
             NULL instead. If multiple ranks provide valid file names, the one
             with the lowest rank in comm will read and distribute
             the coupling configuration.
*/
    void yac_cinit_comm_group_instance (
      const char * xml_filename, const char * schema_filename, MPI_Comm comm,
      const char * group_name, int * yac_instance_id);

/** Elementary initialisation of the whole system

     @param[in]  couple_config_id id of a coupling configuration
     @remark In case the user has multiple YAC instances in parallel, he has
             to initialise YAXT himself.
     @remark In case a process in MPI_COMM_WORLD does not wish to initialise
             YAC, it can call \ref yac_cinit_dummy instead.
     @remark At least one process has to provide valid id for
             couple_config_id. All other processes can provide
             \ref COUPLE_CONFIG_DUMMY_ID instead. If multiple ranks
             provide valid coupling configuration, the one
             with the lowest rank in MPI_COMM_WORLD will read and distribute
             the coupling configuration.
*/
    void yac_cinit_couple_config( int couple_config_id );

/** Elementary initialisation of the whole system

     @param[in]  couple_config_id id of a coupling configuration
     @param[in]  group_name      name of a group of processes
                                 (NULL is a valid argument)
     @remark In case the user has multiple YAC instances in parallel, he has
             to initialise YAXT himself.
     @remark In case a process in MPI_COMM_WORLD does not wish to initialise
             YAC, it can call \ref yac_cinit_dummy instead.
     @remark YAC will generate a communicator for all processes in
             MPI_COMM_WORLD that provid the same non-NULL group name.
             This communicator can be retrieved by
             \ref yac_cget_groupcomm_instance.
     @remark At least one process has to provide valid id for
             couple_config_id. All other processes can provide
             \ref COUPLE_CONFIG_DUMMY_ID instead. If multiple ranks
             provide valid coupling configuration, the one
             with the lowest rank in MPI_COMM_WORLD will read and distribute
             the coupling configuration.
*/
    void yac_cinit_couple_config_group( int couple_config_id,
                                        const char * group_name );

/** Elementary initialisation of the whole system

     @param[in]  couple_config_id id of a coupling configuration
     @param[out] yac_instance_id id of the YAC instance initialised by
                                 this call
     @remark In case the user has multiple YAC instances in parallel, he has
             to initialise YAXT himself.
     @remark In case a process in MPI_COMM_WORLD does not wish to initialise
             YAC, it can call \ref yac_cinit_dummy instead.
     @remark At least one process has to provide valid id for
             couple_config_id. All other processes can provide
             \ref COUPLE_CONFIG_DUMMY_ID instead. If multiple ranks
             provide valid coupling configuration, the one
             with the lowest rank in MPI_COMM_WORLD will read and distribute
             the coupling configuration.
*/
    void yac_cinit_couple_config_instance( int couple_config_id,
                                           int * yac_instance_id );

/** Elementary initialisation of the whole system

     @param[in]  couple_config_id id of a coupling configuration
     @param[in]  group_name      name of a group of processes
                                 (NULL is a valid argument)
     @param[out] yac_instance_id id of the YAC instance initialised by
                                 this call
     @remark In case the user has multiple YAC instances in parallel, he has
             to initialise YAXT himself.
     @remark In case a process in MPI_COMM_WORLD does not wish to initialise
             YAC, it can call \ref yac_cinit_dummy instead.
     @remark YAC will generate a communicator for all processes in
             MPI_COMM_WORLD that provid the same non-NULL group name.
             This communicator can be retrieved by
             \ref yac_cget_groupcomm_instance.
     @remark At least one process has to provide valid id for
             couple_config_id. All other processes can provide
             \ref COUPLE_CONFIG_DUMMY_ID instead. If multiple ranks
             provide valid coupling configuration, the one
             with the lowest rank in MPI_COMM_WORLD will read and distribute
             the coupling configuration.
*/
    void yac_cinit_couple_config_group_instance( int couple_config_id,
                                                 const char * group_name,
                                                 int * yac_instance_id );

/** Elementary initialisation of the whole system using a user-provided
 *  MPI communicator

     @param[in]  couple_config_id id of a coupling configuration
     @param[in]  comm            MPI communicator
     @remark In case the user has multiple YAC instances in parallel, he has
             to initialise YAXT himself.
     @remark In case a process in comm does not wish to initialise
             YAC, it can call \ref yac_cinit_comm_dummy instead.
     @remark At least one process has to provide valid id for
             couple_config_id. All other processes can provide
             \ref COUPLE_CONFIG_DUMMY_ID instead. If multiple ranks
             provide valid coupling configuration, the one
             with the lowest rank in comm will read and distribute
             the coupling configuration.
*/
    void yac_cinit_couple_config_comm( int couple_config_id, MPI_Comm comm );

/** Elementary initialisation of the whole system using a user-provided
 *  MPI communicator

     @param[in]  couple_config_id id of a coupling configuration
     @param[in]  comm            MPI communicator
     @param[in]  group_name      name of a group of processes
                                 (NULL is a valid argument)
     @remark In case the user has multiple YAC instances in parallel, he has
             to initialise YAXT himself.
     @remark In case a process in comm does not wish to initialise
             YAC, it can call \ref yac_cinit_comm_dummy instead.
     @remark YAC will generate a communicator for all processes in comm that
             provid the same non-NULL group name. This communicator can be
             retrieved by \ref yac_cget_groupcomm_instance.
     @remark At least one process has to provide valid id for
             couple_config_id. All other processes can provide
             \ref COUPLE_CONFIG_DUMMY_ID instead. If multiple ranks
             provide valid coupling configuration, the one
             with the lowest rank in comm will read and distribute
             the coupling configuration.
*/
    void yac_cinit_couple_config_comm_group( int couple_config_id,
                                             MPI_Comm comm,
                                             const char * group_name );

/** Elementary initialisation of the whole system using a user-provided
 *  MPI communicator

     @param[in]  couple_config_id id of a coupling configuration
     @param[in]  comm            MPI communicator
     @param[out] yac_instance_id id of the YAC instance initialised by
                                 this call
     @remark In case the user has multiple YAC instances in parallel, he has
             to initialise YAXT himself.
     @remark In case a process in comm does not wish to initialise
             YAC, it can call \ref yac_cinit_comm_dummy instead.
     @remark At least one process has to provide valid id for
             couple_config_id. All other processes can provide
             \ref COUPLE_CONFIG_DUMMY_ID instead. If multiple ranks
             provide valid coupling configuration, the one
             with the lowest rank in comm will read and distribute
             the coupling configuration.
*/
    void yac_cinit_couple_config_comm_instance( int couple_config_id,
                                                MPI_Comm comm,
                                                int * yac_instance_id );

/** Elementary initialisation of the whole system using a user-provided
 *  MPI communicator

     @param[in]  couple_config_id id of a coupling configuration
     @param[in]  comm            MPI communicator
     @param[in]  group_name      name of a group of processes
                                 (NULL is a valid argument)
     @param[out] yac_instance_id id of the YAC instance initialised by
                                 this call
     @remark In case the user has multiple YAC instances in parallel, he has
             to initialise YAXT himself.
     @remark In case a process in comm does not wish to initialise
             YAC, it can call \ref yac_cinit_comm_dummy instead.
     @remark YAC will generate a communicator for all processes in comm that
             provid the same non-NULL group name. This communicator can be
             retrieved by \ref yac_cget_groupcomm_instance.
     @remark At least one process has to provide valid id for
             couple_config_id. All other processes can provide
             \ref COUPLE_CONFIG_DUMMY_ID instead. If multiple ranks
             provide valid coupling configuration, the one
             with the lowest rank in comm will read and distribute
             the coupling configuration.
*/
    void yac_cinit_couple_config_comm_group_instance( int couple_config_id,
                                                      MPI_Comm comm,
                                                      const char * group_name,
                                                      int * yac_instance_id );

/* -------------------------------------------------------------------------------- */

/** Dummy for initialisation of the whole system
     @remark This routine can be called instead of \ref yac_cinit or
             \ref yac_cinit_instance, if the local process does not wish to
             initialise a YAC instance.
*/
    void yac_cinit_dummy();

/** Dummy initialisation of the whole system using a user-provided
 *  MPI world communicator

     @param[in] comm MPI communicator
     @remark This routine can be called instead of \ref yac_cinit_comm or
             \ref yac_cinit_comm_instance, if the local process does not wish to
             initialise a YAC instance.
*/
    void yac_cinit_comm_dummy ( MPI_Comm comm );

/* -------------------------------------------------------------------------------- */

/** Clean-up YAC (same as yac_cfinalize but without MPI_Finalize; see \ref CRestart)
     @remark In case the user initialised yaxt himself, he has to finalise it after
             yac_ccleanup in order to be able to call yac_cinit again.
     @remark cleans up default YAC instance
*/
     void yac_ccleanup ();

/** Clean-up YAC (same as yac_cfinalize_instance but without MPI_Finalize;
 *  see \ref CRestart)
     @param[in] yac_instance_id id of the YAC instance to be cleaned up
     @remark In case the user initialised yaxt himself, he has to finalise it after
             yac_ccleanup in order to be able to call yac_cinit again.
*/
     void yac_ccleanup_instance ( int yac_instance_id );

/* -------------------------------------------------------------------------------- */

/** Cleans up default YAC instance (if available) and finales MPI,
 *  if it was initialised by YAC
*/
     void yac_cfinalize ();

/** Cleans up provided YAC instance and finales MPI,
 *  if it was initialised by YAC
     @param[in] yac_instance_id id of the YAC instance to be finalised
*/
     void yac_cfinalize_instance (int yac_instance_id);

/* -------------------------------------------------------------------------------- */

/** Definition of job start and end datetime for the default YAC instance

     @param[in] start_datetime calendar job start datetime
     @param[in] end_datetime   calendar job end datetime
 */

     void yac_cdef_datetime ( const char * start_datetime,
                              const char * end_datetime );

/** Definition of job start and end datetime

     @param[in] yac_instance_id id of the YAC instance
     @param[in] start_datetime  calendar job start datetime
     @param[in] end_datetime    calendar job end datetime
 */

     void yac_cdef_datetime_instance ( int yac_instance_id,
                                       const char * start_datetime,
                                       const char * end_datetime );

/* -------------------------------------------------------------------------------- */

/** Providing a group communicator for the default YAC instance

     @param[out] group_comm communicator containing all processes that
                            passed the same group name to the init-routine
                            or MPI_COMM_NULL if not group name was provided.
 */

     void yac_cget_groupcomm( MPI_Comm * group_comm );

/** Providing a group communicator

     @param[in] yac_instance_id id of the YAC instance
     @param[out] group_comm communicator containing all processes that
                            passed the same group name to the init-routine
                            or MPI_COMM_NULL if not group name was provided.
 */

     void yac_cget_groupcomm_instance( int yac_instance_id,
                                       MPI_Comm * group_comm );

/* -------------------------------------------------------------------------------- */

/** Elementary definition of the component for the default YAC instance

     @param[in]  comp_name name of the component
     @param[out] comp_id   component Id
     @remark this call is collective for all processes that initialised the
             default YAC instance
     @remark components can only be defined once in the initialisation phase of
             a YAC instance
*/

     void yac_cdef_comp ( const char * comp_name,
                          int * comp_id );

/** Elementary definition of the component.

     @param[in]  yac_instance_id id of the YAC instance
     @param[in]  comp_name name of the component
     @param[out] comp_id   component Id
     @remark this call is collective for all processes that initialised the
             provided YAC instance
     @remark components can only be defined once in the initialisation phase of
             a YAC instance
*/

     void yac_cdef_comp_instance ( int yac_instance_id,
                                   const char * comp_name,
                                   int * comp_id );

/* -------------------------------------------------------------------------------- */

/** Elementary definition of the components for the default YAC instance

     @param[in]  comp_names names of the components
     @param[in]  num_comps  number of components
     @param[out] comp_ids   component Id's
     @remark this call is collective for all processes that initialised the
             default YAC instance
     @remark components can only be defined once in the initialisation phase of
             a YAC instance
*/

     void yac_cdef_comps ( const char ** comp_names,
                           int num_comps,
                           int * comp_ids );

/** Elementary definition of the components.

     @param[in]  yac_instance_id id of the YAC instance
     @param[in]  comp_names names of the components
     @param[in]  num_comps  number of components
     @param[out] comp_ids   component Id's
     @remark this call is collective for all processes that initialised the
             provided YAC instance
     @remark components can only be defined once in the initialisation phase of
             a YAC instance
*/

     void yac_cdef_comps_instance ( int yac_instance_id,
                                    const char ** comp_names,
                                    int num_comps,
                                    int * comp_ids );

/* -------------------------------------------------------------------------------- */

/** Setting io_proc_step for a component.

     @param[in] comp_id      - component Id
     @param[in] io_proc_step - stride between component processes doing io
                               (e.g. writing weight files)
*/

     void yac_cset_comp_io_proc_step ( int comp_id,
                                       int io_proc_step );

/* -------------------------------------------------------------------------------- */

/** Getting io_proc_step from a component.

     @param[in] comp_id       - component Id
     @param[out] io_proc_step - stride between component processes doing io
                                (e.g. writing weight files)
*/

     void yac_cget_comp_io_proc_step ( int comp_id,
                                       int * io_proc_step );

/* -------------------------------------------------------------------------------- */

/** Providing the local MPI communicator

     @param[in]  comp_id            - local component ID (from yac_cdef_comp), handle to the component struct
     @param[out] local_communicator - returned local communicator

*/
     void yac_cget_localcomm ( MPI_Comm *local_communicator,
                               int comp_id );

/* -------------------------------------------------------------------------------- */

/** Providing a MPI communicator for communication between the component
 *  root processes

     @param[out] local_communicator returned local communicator
     @param[in]  comp_name1         name of the first component
     @param[in]  comp_name2         name of the second component
     @remark the components have to defined in the default YAC instance
*/
     void yac_cget_pair_rootcomm ( MPI_Comm *local_communicator,
                                   const char * comp_name1,
                                   const char * comp_name2 );

/** Providing a MPI communicator for communication between the component
 *  root processes

     @param[in]  yac_instance_id    id of the YAC instance
     @param[out] local_communicator returned local communicator
     @param[in]  comp_name1         name of the first component
     @param[in]  comp_name2         name of the second component
*/
     void yac_cget_pair_rootcomm_instance ( int yac_instance_id,
                                            MPI_Comm *local_communicator,
                                            const char * comp_name1,
                                            const char * comp_name2 );

/* -------------------------------------------------------------------------------- */

/** Definition of a set of points for 2D regular grids.

     @param[in]  grid_id    handle to the grid struct (returned from yac_cdef_grid)
     @param[in]  nbr_points number of points
     @param[in]  location   location of points [edge|vertex|center]
     @param[in]  x_points   array of point longitudes
     @param[in]  y_points   array of point latitudes
     @param[out] point_id   returned handle to the points struct
*/

     void yac_cdef_points_reg2d ( int const grid_id,
                                  int const *nbr_points,
                                  int const location,
                                  double const *x_points,
                                  double const *y_points,
                                  int *point_id );

/* -------------------------------------------------------------------------------- */

/** Definition of a set of points for 2D curvilinear grids.

     @param[in]  grid_id    handle to the grid struct (returned from yac_cdef_grid)
     @param[in]  nbr_points number of points
     @param[in]  location   location of points [edge|vertex|center]
     @param[in]  x_points   array of point longitudes
     @param[in]  y_points   array of point latitudes
     @param[out] point_id   returned handle to the points struct
*/

     void yac_cdef_points_curve2d ( int const grid_id,
                                    int const *nbr_points,
                                    int const location,
                                    double const *x_points,
                                    double const *y_points,
                                    int *point_id );

/* -------------------------------------------------------------------------------- */

/** Definition of a set of points for 2D unstructured grids.

     @param[in]  grid_id    handle to the grid struct (returned from yac_cdef_grid)
     @param[in]  nbr_points number of points
     @param[in]  location   location of points [edge|vertex|center]
     @param[in]  x_points   array of point longitudes
     @param[in]  y_points   array of point latitudes
     @param[out] point_id   returned handle to the points struct
*/

     void yac_cdef_points_unstruct ( int const grid_id,
                                     int const nbr_points,
                                     int const location,
                                     double const *x_points,
                                     double const *y_points,
                                     int *point_id );

/* -------------------------------------------------------------------------------- */

/** Definition of a 2d regular grid

     @param[in]  grid_name    name of the grid
     @param[in]  nbr_vertices 2d array containing the number of vertices in each dimension
     @param[in]  cyclic       2d array containing information about cyclic behaviour in each dimension
     @param[in]  x_vertices   array of vertex longitudes
     @param[in]  y_vertices   array of vertex latitudes
     @param[out] grid_id      id of generated grid
     @remark on the local order of edges:
             generate for each edge a tuple containing the local indices of the
             adjacent vertices (lowest index first) and sort this list by the
             vertex indices; this gives the the local order of edges
*/
     void yac_cdef_grid_reg2d ( const char * grid_name,
                                int nbr_vertices[2],
                                int cyclic[2],
                                double *x_vertices,
                                double *y_vertices,
                                int *grid_id);

/* -------------------------------------------------------------------------------- */

/** Definition of a 2d curvilinear grid

     @param[in]  grid_name    name of the grid
     @param[in]  nbr_vertices 2d array containing the number of vertices in each dimension
     @param[in]  cyclic       2d array containing information about cyclic behaviour in each dimension
     @param[in]  x_vertices   array of vertex longitudes
     @param[in]  y_vertices   array of vertex latitudes
     @param[out] grid_id      id of generated grid
     @remark on the local order of edges:
             generate for each edge a tuple containing the local indices of the
             adjacent vertices (lowest index first) and sort this list by the
             vertex indices; this gives the the local order of edges
*/
     void yac_cdef_grid_curve2d ( const char * grid_name,
                                  int nbr_vertices[2],
                                  int cyclic[2],
                                  double *x_vertices,
                                  double *y_vertices,
                                  int *grid_id);

/* -------------------------------------------------------------------------------- */

/** Definition of an unstructured grid

     @param[in]  grid_name             name of the grid
     @param[in]  nbr_vertices          number of vertices
     @param[in]  nbr_cells             number of cells
     @param[in]  num_vertices_per_cell array containing the number of vertices for each cell
     @param[in]  x_vertices            array of vertex longitudes
     @param[in]  y_vertices            array of vertex latitudes
     @param[in]  cell_to_vertex        connectivity of vertices belonging to cells\n
                                       (the vertex indices per cell have to be in clockwise or counterclockwise ordering)
     @param[out] grid_id               id of generated grid
*/
     void yac_cdef_grid_unstruct ( const char * grid_name,
                                   int nbr_vertices,
                                   int nbr_cells,
                                   int *num_vertices_per_cell,
                                   double *x_vertices,
                                   double *y_vertices,
                                   int *cell_to_vertex,
                                   int *grid_id);

/* -------------------------------------------------------------------------------- */

/** Set global ids for a grid

     @param[in] global_index array of global indices
     @param[in] location     cell/vertex/edge
     @param[in] grid_id      grid id
     @remark global indices are to be provided in the local order of their
             respective cells/vertices/edges
*/

     void yac_cset_global_index ( int const * global_index,
                                  int location,
                                  int grid_id);

/* -------------------------------------------------------------------------------- */

/** Set core mask for a grid

     @param[in] is_core   0 for cells/vertices/edges that are halos,
                          1 for cells/vertices/edges that are core
     @param[in] location  cell/vertex/edge
     @param[in] grid_id   grid id
     @remark cells/vertices/edges who are halos are not used or set
             in a put or get operation
     @remark core mask values are to be provided in the local order of their
             respective cells/vertices/edges
*/

     void yac_cset_core_mask ( int const * is_core,
                               int location,
                               int grid_id);

/* -------------------------------------------------------------------------------- */

/** Set the default mask for points

     @param[in] is_valid   0 for points that are masked out, 1 for valid points
     @param[in] points_id  id of points/cells
*/

     void yac_cset_mask ( int const * is_valid,
                          int points_id );

/* -------------------------------------------------------------------------------- */

/** define a mask for a grid

     @param[in]  grid_id    grid ID
     @param[in]  nbr_points number of points
     @param[in]  location   cell/vertex/edge
     @param[in]  is_valid   0 for points that are masked out, 1 for valid points
     @param[out] mask_id    returned handle to the mask struct
*/

     void yac_cdef_mask ( int const grid_id,
                          int const nbr_points,
                          int const location,
                          int const * is_valid,
                          int *mask_id );

/* -------------------------------------------------------------------------------- */

/** Definition of the coupling field (using default mask, if defined)

     @param[in]  field_name               - character string providing the name of the coupling field
     @param[in]  component_id             - component ID
     @param[in]  point_ids                - point IDs
     @param[in]  num_pointsets            - number of pointsets per grid
     @param[out] field_id                 - returned field_id which has to be used to identify coupling fields
                                            in yac_cput and yac_cget.
*/

       void yac_cdef_field ( char const * field_name,
                             int const component_id,
                             int const * point_ids,
                             int const num_pointsets,
                             int * field_id );

/* -------------------------------------------------------------------------------- */

/** Definition of the coupling field

     @param[in]  field_name               - character string providing the name of the coupling field
     @param[in]  component_id             - component ID
     @param[in]  point_ids                - point IDs
     @param[in]  mask_ids                 - mask IDs
     @param[in]  num_pointsets            - number of pointsets per grid
     @param[out] field_id                 - returned field_id which has to be used to identify coupling fields
                                            in yac_cput and yac_cget.
*/

       void yac_cdef_field_mask ( char const * field_name,
                                  int const component_id,
                                  int const * point_ids,
                                  int const * mask_ids,
                                  int const num_pointsets,
                                  int * field_id );

/* -------------------------------------------------------------------------- */

/** Returns the action a put/get would return for the current timestep

@param[in]  field_id field id returned by \ref yac_cdef_field
@param[out] action   action for the current timestep
                          (\ref YAC_ACTION_NONE,
                           \ref YAC_ACTION_REDUCTION,
                           \ref YAC_ACTION_COUPLING,
                           \ref YAC_ACTION_GET_FOR_RESTART,
                           \ref YAC_ACTION_PUT_FOR_RESTART, or
                           \ref YAC_ACTION_OUT_OF_BOUND)
@remark If the exchange type of the field is \ref YAC_EXCHANGE_TYPE_NONE,
        this routine will return \ref YAC_ACTION_NONE.
@remark If this routine is called before \ref yac_csearch, the behaviour
        is undefined.
*/

void yac_cget_action(int field_id, int * action);

/* -------------------------------------------------------------------------- */

/** If the action for the current timestep is \ref YAC_ACTION_NONE,
    this routine will advance the internal clock to next timestep.

@param[in]  field_id field id returned by \ref yac_cdef_field
@remark If this routine is called before \ref yac_csearch, the behaviour
        is undefined.
@remark This routine can be called instead of \ref yac_cput or \ref yac_cget,
        if these routines would not have taken any action for the current
        timestep, in order to advance the internal clock.
*/

void yac_cupdate(int field_id);

/* -------------------------------------------------------------------------- */

/** Receiving of the coupling fields

     @param[in]  field_id           -
     @param[in]  collection_size    -
     @param[in]  recv_field         - receive buffer (all data is stored in one
                                                      contiguous part of the memory)
                                      dimensions: recv_field[collection_size][nbr_points]
     @param[in]  info               - returned info argument indicating the action performed
     @param[out] ierror             - returned error
*/
     void yac_cget_ ( int const field_id,
                      int const collection_size,
                      double *recv_field,
                      int    *info,
                      int    *ierror );

/** Receiving of the coupling fields

     @param[in]  field_id           -
     @param[in]  collection_size    -
     @param[in]  recv_field         - receive buffer
                                      dimensions: recv_field[collection_size][nbr_points]
     @param[out] info               - returned info argument indicating the action performed
     @param[out] ierror             - returned error
*/
     void yac_cget ( int const field_id,
                     int const collection_size,
                     double **recv_field,
                     int    *info,
                     int    *ierror );

/* -------------------------------------------------------------------------------- */

/** Sending of the coupling fields

     @param[in]  field_id           -
     @param[in]  collection_size    -
     @param[in]  send_field         - send buffer (all data is stored in one
                                                   contiguous part of the memory)
                                      dimensions send_field[collection_size]
                                                           [nbr_fields]
                                                           [nbr_points]
     @param[out] info               - returned info
     @param[out] ierror             - returned error
*/

     void yac_cput_ ( int const field_id,
                      int const collection_size,
                      double *send_field ,
                      int    *info,
                      int    *ierror );

/** Sending of the coupling fields

     @param[in]  field_id           -
     @param[in]  collection_size    -
     @param[in]  send_field         - send buffer
                                      dimensions send_field[collection_size]
                                                           [nbr_fields]
                                                           [nbr_points]
     @param[out] info               - returned info
     @param[out] ierror             - returned error
*/

     void yac_cput ( int const field_id,
                     int const collection_size,
                     double *** const send_field,
                     int *info,
                     int *ierror );

/* -------------------------------------------------------------------------------- */

/** Sending of the coupling fields

     @param[in]  field_id           -
     @param[in]  collection_size    -
     @param[in]  send_field         - send buffer
                                      dimensions send_field[collection_size * nbr_fields]
                                                           [nbr_points]
     @param[out] info               - returned info
     @param[out] ierror             - returned error
*/

     void yac_cput_ptr_ ( int const field_id,
                          int const collection_size,
                          double ** send_field,
                          int    *info,
                          int    *ierror );

/* -------------------------------------------------------------------------------- */

/** Exchange of the coupling fields

     @param[in]  send_field_id      -
     @param[in]  recv_field_id      -
     @param[in]  collection_size    -
     @param[in]  send_field         - send buffer (all data is stored in one
                                                   contiguous part of the memory)
                                      dimensions send_field[collection_size]
                                                           [nbr_fields]
                                                           [nbr_points]
     @param[in]  recv_field         - receive buffer (all data is stored in one
                                                      contiguous part of the memory)
                                      dimensions: recv_field[collection_size][nbr_points]
     @param[out] send_info          - returned send info
     @param[out] recv_info          - returned recv info
     @param[out] ierror             - returned error
*/

     void yac_cexchange_ ( int const send_field_id,
                           int const recv_field_id,
                           int const collection_size,
                           double *send_field ,
                           double *recv_field,
                           int    *send_info,
                           int    *recv_info,
                           int    *ierror );


/** Exchange of the coupling fields

     @param[in]  send_field_id      -
     @param[in]  recv_field_id      -
     @param[in]  collection_size    -
     @param[in]  send_field         - send buffer
                                      dimensions send_field[collection_size]
                                                           [nbr_fields]
                                                           [nbr_points]
     @param[in]  recv_field         - receive buffer
                                      dimensions: recv_field[collection_size][nbr_points]
     @param[out] send_info          - returned send info
     @param[out] recv_info          - returned recv info
     @param[out] ierror             - returned error
*/

     void yac_cexchange ( int const send_field_id,
                          int const recv_field_id,
                          int const collection_size,
                          double *** const send_field,
                          double ** recv_field,
                          int *send_info,
                          int *recv_info,
                          int *ierror );

/* -------------------------------------------------------------------------------- */

/** Sending of the coupling fields

     @param[in]  send_field_id      -
     @param[in]  recv_field_id      -
     @param[in]  collection_size    -
     @param[in]  send_field         - send buffer
                                      dimensions send_field[collection_size * nbr_fields]
                                                           [nbr_points]
     @param[in]  recv_field         - receive buffer
                                      dimensions: recv_field[collection_size][nbr_points]
     @param[out] send_info          - returned send info
     @param[out] recv_info          - returned recv info
     @param[out] ierror             - returned error
*/

     void yac_cexchange_ptr_ ( int const send_field_id,
                               int const recv_field_id,
                               int const collection_size,
                               double ** send_field,
                               double ** recv_field,
                               int    *send_info,
                               int    *recv_info,
                               int    *ierror );

/* -------------------------------------------------------------------------------- */

/** End of the definition phase for the default YAC instance,
 *  invocation of the search

     @param[out] ierror returned error
*/

void yac_csearch ( int * ierror );

/** End of the definition phase, invocation of the search

     @param[in]  yac_instance_id id of the YAC instance
     @param[out] ierror          returned error
*/

void yac_csearch_instance ( int yac_instance_id, int * ierror );

/* --------------------------------------------------------------------------------
           query routines
   -------------------------------------------------------------------------------- */

/** query routine for the start datetime of the job of the default YAC instance
*/

char * yac_cget_start_datetime ();

/** query routine for the start datetime of the job

     @param[in] yac_instance_id id of the YAC instance
*/

char * yac_cget_start_datetime_instance ( int yac_instance_id );

/** query routine for the end datetime of the job of the default YAC instance
*/

char * yac_cget_end_datetime ();

/** query routine for the end datetime of the job

     @param[in] yac_instance_id id of the YAC instance
*/

char * yac_cget_end_datetime_instance ( int yac_instance_id );

/** query routine for the yac version
*/

char * yac_cget_version ();

/* -------------------------------------------------------------------------------- */

/** query routine for number of coupling fields defined of the default YAC
 *  instance

     @param[in] nbr_fields         - number of fields defined on the calling process

*/

void yac_cget_nbr_fields ( int *nbr_fields );

/** query routine for number of coupling fields defined

     @param[in] yac_instance_id id of the YAC instance
     @param[in] nbr_fields      number of fields defined on the calling process

*/

void yac_cget_nbr_fields_instance ( int yac_instance_id, int *nbr_fields );

/* -------------------------------------------------------------------------------- */

/** query routine to get the list of field IDs for defined fields of the
 *  default YAC instance

     @param[in] nbr_fields number of fields defined on the calling process
     @param[in] field_ids  list of delivered field IDs

*/

void yac_cget_field_ids ( int nbr_fields, int *field_ids );

/** query routine to get the list of field IDs for defined fields

     @param[in] yac_instance_id id of the YAC instance
     @param[in] nbr_fields number of fields defined on the calling process
     @param[in] field_ids  list of delivered field IDs

*/

void yac_cget_field_ids_instance ( int yac_instance_id,
                                   int nbr_fields,
                                   int * field_ids );

/* ---------------------------------------------------------------------- */

/** query routine to get the field ID for a given name of the default YAC
 *  instance

     @param[in] field_name field name that has been given to yac_cdef_field
     @param[in] field_id   ID as it has been provided by yac_cdef_field
     @remark in case a field is define multiple times on the local processes
             (for different grids) use
             \ref yac_cget_field_id_from_field_and_grid_name
             to get the field ID

*/

void yac_cget_field_id_from_name ( const char *field_name, int *field_id );

/** query routine to get the field ID for a given name of the default YAC
 *  instance

     @param[in] yac_instance_id id of the YAC instance
     @param[in] field_name      field name that has been given to yac_cdef_field
     @param[in] field_id        ID as it has been provided by yac_cdef_field
     @remark in case a field is define multiple times on the local processes
             (for different grids) use
             \ref yac_cget_field_id_from_field_and_grid_name_instance
             to get the field ID

*/

void yac_cget_field_id_from_name_instance ( int yac_instance_id,
                                            const char *field_name,
                                            int *field_id );

/* ---------------------------------------------------------------------- */

/** query routine to get the field name for a given ID of the default YAC
 *  instance

     @param[in] field_name field name that has been given to yac_cdef_field
     @param[in] grid_name  grid name for which the field has been defined
     @param[in] field_id   ID as it has been provided by yac_cdef_field

*/

void yac_cget_field_id_from_field_and_grid_name(
  const char *field_name, const char *grid_name, int *field_id);

/** query routine to get the field name for a given ID of the default YAC
 *  instance

     @param[in] yac_instance_id id of the YAC instance
     @param[in] field_name      field name that has been given to yac_cdef_field
     @param[in] grid_name       grid name for which the field has been defined
     @param[in] field_id        ID as it has been provided by yac_cdef_field

*/

void yac_cget_field_id_from_field_and_grid_name_instance(
  int yac_instance_id, const char *field_name, const char *grid_name,
  int *field_id );

/* ---------------------------------------------------------------------- */

/** query routine to get the field name for a given ID of the default YAC
 *  instance

     @param[in] field_id   ID as provided by yac_cdef_field
     @param[in] field_name field name that has been given to yac_cdef_field

*/

void yac_cget_field_name_from_id ( int field_id, char *field_name );

/* ---------------------------------------------------------------------- */

/** query routine to get the requested send/recv operation for a given field
 *  name of the default YAC instance

     @param[in] field_name field name as provided by yac_cdef_field
     @param[in] role       exchange type of the field\n
                           (\ref YAC_EXCHANGE_TYPE_NONE,
                            \ref YAC_EXCHANGE_TYPE_SOURCE, or
                            \ref YAC_EXCHANGE_TYPE_TARGET)
     @remark the role is set by the call to \ref yac_csearch or
             \ref yac_csearch_instance
     @remark in case a field is define multiple times on the local processes
             (for different grids) use
             \ref yac_cget_role_of_field_by_field_and_grid_name
             to get the role

*/

void yac_cget_role_of_field_by_name ( char const * field_name, int *role );

/** query routine to get the requested send/recv operation for a given field
  * name

     @param[in] yac_instance_id id of the YAC instance
     @param[in] field_name      field name as provided by yac_cdef_field
     @param[in] role            exchange type of the field\n
                                (\ref YAC_EXCHANGE_TYPE_NONE,
                                 \ref YAC_EXCHANGE_TYPE_SOURCE, or
                                 \ref YAC_EXCHANGE_TYPE_TARGET)

     @remark the role is set by the call to \ref yac_csearch or
             \ref yac_csearch_instance
     @remark in case a field is define multiple times on the local processes
             (for different grids) use
             \ref yac_cget_role_of_field_by_field_and_grid_name_instance
             to get the role

*/

void yac_cget_role_of_field_by_name_instance ( int yac_instance_id,
                                               char const * field_name,
                                               int *role );

/* ---------------------------------------------------------------------- */

/** query routine to get the requested send/recv operation for a given field ID
 *  of the default YAC instance

     @param[in] field_name field name as provided by yac_cdef_field
     @param[in] grid_name  grid name for which the field has been defined
     @param[in] role       exchange type of the field\n
                           (\ref YAC_EXCHANGE_TYPE_NONE,
                            \ref YAC_EXCHANGE_TYPE_SOURCE, or
                            \ref YAC_EXCHANGE_TYPE_TARGET)
     @remark the role is set by the call to \ref yac_csearch or
             \ref yac_csearch_instance

*/

void yac_cget_role_of_field_by_field_and_grid_name (
  char const * field_name, char const * grid_name, int *role );

/** query routine to get the requested send/recv operation for a given field ID

     @param[in] yac_instance_id id of the YAC instance
     @param[in] field_name      field name as provided by yac_cdef_field
     @param[in] grid_name       grid name for which the field has been defined
     @param[in] role            exchange type of the field\n
                                (\ref YAC_EXCHANGE_TYPE_NONE,
                                 \ref YAC_EXCHANGE_TYPE_SOURCE, or
                                 \ref YAC_EXCHANGE_TYPE_TARGET)
     @remark the role is set by the call to \ref yac_csearch or
             \ref yac_csearch_instance

*/

void yac_cget_role_of_field_by_field_and_grid_name_instance (
  int yac_instance_id, char const * field_name, char const * grid_name,
  int *role );

/* ---------------------------------------------------------------------- */

/** query routine to get the requested send/recv operation for a given field ID

     @param[in] field_id ID as provided by yac_cdef_field
     @param[in] role     exchange type of the field\n
                         (\ref YAC_EXCHANGE_TYPE_NONE,
                          \ref YAC_EXCHANGE_TYPE_SOURCE, or
                          \ref YAC_EXCHANGE_TYPE_TARGET)

*/

void yac_cget_role_of_field_by_id ( int field_id, int *role );

/* ---------------------------------------------------------------------- */

/** query routine to get coupling interval for a given field name of the default
 *  YAC instance

     @param[in] field_name      field name as set in yac_cdef_field
     @param[in] timestep_string coupling time step as iso8601 string

*/

void yac_cget_coupling_timestep_name ( const char * field_name,
                                       char *timestep_string );

/** query routine to get coupling interval for a given field name

     @param[in] yac_instance_id id of the YAC instance
     @param[in] field_name      field name as set in yac_cdef_field
     @param[in] timestep_string coupling time step as iso8601 string

*/

void yac_cget_coupling_timestep_name_instance ( int yac_instance_id,
                                                const char * field_name,
                                                char *timestep_string );

/* ---------------------------------------------------------------------- */

/** query routine to get coupling interval for a given field id

     @param[in] field_id        field ID (as representative for a field name)
     @param[in] timestep_string coupling time step as iso8601 string

*/

void yac_cget_coupling_timestep_id ( const int field_id,
                                     char *timestep_string );

/* ---------------------------------------------------------------------- */

/** query routine to get collection size for a given field name of the default
 *  YAC instance

     @param[in] field_name      field name as set in yac_cdef_field
     @param[in] collection_size collection size

*/

void yac_cget_collection_size_name ( const char * field_name,
                                     int * collection_size );

/** query routine to get collection size for a given field name

     @param[in] yac_instance_id id of the YAC instance
     @param[in] field_name      field name as set in yac_cdef_field
     @param[in] collection_size collection size

*/

void yac_cget_collection_size_name_instance ( int yac_instance_id,
                                              const char * field_name,
                                              int * collection_size );

/* ---------------------------------------------------------------------- */

/** query routine to get collection size for a given field id

     @param[in] field_id        field ID (as representative for a field name)
     @param[in] collection_size collection size

*/

void yac_cget_collection_size_id ( const int field_id,
                                   int * collection_size );

/* -------------------------------------------------------------------------------- */

/** query routine to get model time step for a given field name of the default
 *  YAC instance

     @param[in] field_name      field name as set in yac_cdef_field
     @param[in] timestep_string model time step as iso8601 string

*/

void yac_cget_model_timestep_name ( const char * field_name,
                                    char *timestep_string );

/** query routine to get model time step for a given field name

     @param[in] yac_instance_id id of the YAC instance
     @param[in] field_name      field name as set in yac_cdef_field
     @param[in] timestep_string model time step as iso8601 string

*/

void yac_cget_model_timestep_name_instance ( int yac_instance_id,
                                             const char * field_name,
                                             char *timestep_string );

/* -------------------------------------------------------------------------------- */

/** query routine to get model time step for a given field id

     @param[in] field_id           - field ID (as representative for a field name)
     @param[in] timestep_string    - model time step as iso8601 string

*/

void yac_cget_model_timestep_id ( int field_id,
                                  char *timestep_string );

/* --------------------------------------------------------------------------------
           auxiliary routines
   -------------------------------------------------------------------------------- */

/**
 * redirect stderr and stdout into files with names depending on the process ID
 *
 * The original version was part of the OASIS4 coupler and has been adjusted to
 * work for yac.
 *
 * @param[in]  filestem   basename of stdout/stderr files
 * @param[in]  parallel   flag for direction of output
                             parallel = 1 : Create single output files for each process
                             otherwise    : Create output files only for process 0.
                                           The output of other processes is not redirected.
 * @param[in]  my_pe      local rank
 * @param[in]  npes       total number of processes in set
 */

void yac_redirstdout ( const char *filestem,
                       int parallel, int my_pe,
                       int npes );

/* -------------------------------------------------------------------------------- */

/** query routine to get the number of points in a grid

     @param[in] location
     @param[in] grid_id

*/

size_t yac_get_grid_size ( int location, int grid_id );

/* -------------------------------------------------------------------------------- */

/** query routine to get the number of points in a pointset

     @param[in] points_id

*/

size_t yac_get_points_size ( int points_id  );

/* -------------------------------------------------------------------------------- */

/** gets the number of active transients for a component

     @param[in] comp_id component id
     @return    number of active transients

*/
int yac_get_transients_count(int comp_id);

/** gets ids for all active transients

     @param[in]  comp_id       component id
     @param[out] transient_ids transient ids
     @remark the user has to provide the array for the transient ids

*/
void yac_get_transient_ids(int comp_id, int * transient_ids);

/** determines whether a transient is source

     @param[in] transient_id transient id
     @return != 0 if the component of the transient is the source
             for this transient
     @remark a transient can be both, source and target, for a component

*/
int yac_is_transient_source(int transient_id);

/** determines whether a transient is target

     @param[in] transient_id transient id
     @return != 0 if the component of the transient is the target
             for this transient
     @remark a transient can be both, source and target, for a component

*/
int yac_is_transient_target(int transient_id);

/** gets the collection size of a transient

     @param[in] transient_id transient id
     @return collection size

*/
int yac_get_transient_collection_size(int transient_id);

/** gets the coupling period of a transient

     @param[in] transient_id transient id
     @return coupling period

*/
char const * yac_get_transient_coupling_period(int transient_id);

/** gets the name of source grid used

     @param[in] transient_id transient id
     @return name of the source grid

*/
char const * yac_get_transient_source_grid_name(int transient_id);

/** gets the name of target grid used

     @param[in] transient_id transient id
     @return name of the target grid

*/
char const * yac_get_transient_target_grid_name(int transient_id);

/** gets the name of a transient

     @param[in] transient_id transient id
     @return name of the transient

*/
char const * yac_get_transient_name(int transient_id);

/* -------------------------------------------------------------------------------- */

/** generates a couple configuration from a given XML configuration file

     @param[in]  xml_filename     XML configuration file
     @param[in]  schema_filename  schema file for the XML configuration file
     @param[out] couple_config_id id to the coupling configuration that was
                                  generated from the XML configuration file
*/
void yac_get_couple_config_xml(
  const char * xml_filename, const char * schema_filename,
  int * couple_config_id);

/** generates an empty couple configuration

     @param[in] couple_config_id id of empty couple configuration
*/
void yac_get_couple_config_empty(int * couple_config_id);

/** frees a couple configuration

     @param[in] couple_config_id id of couple configuration
*/
void yac_free_couple_config(int couple_config_id);

/** sets the timestep unit for a given couple configuration
     @param[in] couple_config_id id of couple configuration
     @param[in] timestep_unit    timestep unit
*/
void yac_set_couple_config_timestep_unit(
  int couple_config_id, int timestep_unit);

/** sets the start and end date for a given couple configuration
     @param[in] couple_config_id id of couple configuration
     @param[in] start            start date
     @param[in] end              end date
*/
void yac_set_couple_config_datetime(
  int couple_config_id, char const * start, char const * end);

/** sets the calendar type for a given couple configuration
     @param[in] couple_config_id id of couple configuration
     @param[in] calendar         calendar type
*/
void yac_set_couple_config_calendar(int couple_config_id, int calendar);

/** adds a grid name to a given couple configuration
     @param[in] couple_config_id id of couple configuration
     @param[in] name             grid name
*/
void yac_add_couple_config_grid(int couple_config_id, char const * name);

/** adds a component name to a given couple configuration
     @param[in] couple_config_id id of couple configuration
     @param[in] name             component name
*/
void yac_add_couple_config_component(int couple_config_id, char const * name);

/** adds a transient to a given couple configuration
     @param[in] couple_config_id id of couple configuration
     @param[in] name             transient name
     @param[in] collection_size  transient collection size
*/
void yac_add_couple_config_transient(
  int couple_config_id, char const * name, size_t collection_size);

/** sets the collection size of a transient of a given couple configuration
     @param[in] couple_config_id id of couple configuration
     @param[in] name             transient name
     @param[in] collection_size  transient collection size
*/
void yac_set_couple_config_transient_collection_size(
  int couple_config_id, char const * name, size_t collection_size);

/** adds transient couple to a given couple configuration
     @param[in] couple_config_id          id of couple configuration
     @param[in] transient_name            transient name
     @param[in] src_component_name        source component name
     @param[in] src_grid_name             source grid name
     @param[in] src_timestep              source time step
     @param[in] src_lag                   source lag
     @param[in] tgt_component_name        target component name
     @param[in] tgt_grid_name             target grid name
     @param[in] tgt_timestep              target time step
     @param[in] tgt_lag                   target lag
     @param[in] mapping_on_source         apply the weights at the
                                          source or target processes
     @param[in] interp_stack_config_id    id of an interpolation stack
     @param[in] coupling_period_operation coupling period operation
     @param[in] coupling_period           coupling period
     @param[in] enforce_write_weight_file write weight file or not
     @param[in] weight_file_name          name of the weight file
*/
void yac_add_couple_config_transient_couple(
  int couple_config_id, char const * transient_name,
  char const * src_component_name, char const * src_grid_name,
  char const * src_timestep, int src_lag,
  char const * tgt_component_name, char const * tgt_grid_name,
  char const * tgt_timestep, int tgt_lag,
  int mapping_on_source, int interp_stack_config_id,
  int coupling_period_operation, char const * coupling_period,
  int enforce_write_weight_file, char const * weight_file_name);

/* -------------------------------------------------------------------------------- */

/** get grid id
     @param[in]  couple_config_id id of couple configuration
     @param[in]  grid_name        grid name
     @param[out] grid_id          grid id
     @remark this routine returns the internal grid id, which is not
             related to the one returned by yac_cdef_grid
*/
void yac_get_couple_config_grid_id(
  int couple_config_id, char const * grid_name, int * grid_id);

/** get component id
     @param[in]  couple_config_id id of couple configuration
     @param[in]  component_name   component name
     @param[out] component_id     component id
     @remark this routine returns the internal component id, which is not
             related to the one returned by yac_cdef_component
*/
void yac_get_couple_config_component_id(
  int couple_config_id, char const * component_name, int * component_id);

/** get transient id
     @param[in]  couple_config_id id of couple configuration
     @param[in]  transient_name   transient name
     @param[out] transient_id     transient id
     @remark this routine returns the internal transient id, which is not
             related to the one returned by yac_cdef_field
*/
void yac_get_couple_config_transient_id(
  int couple_config_id, char const * transient_name, int * transient_id);

/* -------------------------------------------------------------------------------- */

/** gets an empty stack trace
     @param[out] interp_stack_config_id interpolation stack
*/
void yac_get_interp_stack_config(int * interp_stack_config_id);

/** frees a stack trace
     @param[in] interp_stack_config_id interpolation stack
*/
void yac_free_interp_stack_config(int interp_stack_config_id);

/** adds average interpolation to the buttom of a interpolation stack
     @param[in] interp_stack_config_id interpolation stack
     @param[in] reduction_type         reduction type
     @param[in] partial_coverage       allow partial coverage
*/
void yac_add_interp_stack_config_average(
  int interp_stack_config_id, int reduction_type, int partial_coverage);

/** adds n-nearest-neighbour interpolation to the buttom of a
    interpolation stack
     @param[in] interp_stack_config_id interpolation stack
     @param[in] type                   reduction type
     @param[in] n                      number of nearest neighbour points
     @param[in] scale                  scale parameter required by some
                                       reduction types
*/
void yac_add_interp_stack_config_nnn(
  int interp_stack_config_id, int type, size_t n, double scale);

/** adds conservative interpolation to the buttom of a interpolation stack
     @param[in] interp_stack_config_id interpolation stack
     @param[in] order                  first or second order
     @param[in] enforced_conserv       enforce local conservation
     @param[in] partial_coverage       allow partial coverage
     @param[in] normalisation          normalisation type
*/
void yac_add_interp_stack_config_conservative(
  int interp_stack_config_id, int order, int enforced_conserv,
  int partial_coverage, int normalisation);

/** adds source point mapping interpolation to the buttom
    of a interpolation stack
     @param[in] interp_stack_config_id interpolation stack
     @param[in] spread_distance        spread distance
     @param[in] max_search_distance    maximum search distance
     @param[in] weight_type            reduction type
*/
void yac_add_interp_stack_config_spmap(
  int interp_stack_config_id, double spread_distance,
  double max_search_distance, int weight_type);

/** adds HCSBB interpolation to the buttom of a interpolation stack
     @param[in] interp_stack_config_id interpolation stack
*/
void yac_add_interp_stack_config_hcsbb(int interp_stack_config_id);

/** adds user weight file interpolation to the buttom
    of a interpolation stack
     @param[in] interp_stack_config_id interpolation stack
     @param[in] filename               weight file name
     @param[in] src_grid_name          name of source grid
     @param[in] tgt_grid_name          name of target grid
     @param[in] io_proc_step           stride between process ranks that do I/O
*/
void yac_add_interp_stack_config_user_file(
  int interp_stack_config_id, char * filename, char * src_grid_name,
  char * tgt_grid_name, int io_proc_step);

/** adds fixed interpolation to the buttom of a interpolation stack
     @param[in] interp_stack_config_id interpolation stack
     @param[in] value                  fixed value
*/
void yac_add_interp_stack_config_fixed(
  int interp_stack_config_id, double value);

/** adds fixed interpolation to the buttom of a interpolation stack
     @param[in] interp_stack_config_id interpolation stack
     @param[in] constructor_key        key provided to
                                       \ref yac_interp_method_check_add_constructor_callback
                                       for a constructor callback routine
     @param[in] do_search_key          key provided to
                                       \ref yac_interp_method_check_add_do_search_callback
                                       for a do_search callback routine
*/
void yac_add_interp_stack_config_check(
  int interp_stack_config_id, char * constructor_key, char * do_search_key);

/** adds creep interpolation to the buttom of a interpolation stack
     @param[in] interp_stack_config_id interpolation stack
     @param[in] creep_distance         creep distance
*/
void yac_add_interp_stack_config_creep(
  int interp_stack_config_id, int creep_distance);

/** adds user callback interpolation to the buttom of a interpolation stack
     @param[in] interp_stack_config_id   interpolation stack
     @param[in] func_compute_weights_key key provided to
                                         \ref yac_add_compute_weights_callback
                                         for a compute_weights callback routine
*/
void yac_add_interp_stack_config_user_callback(
  int interp_stack_config_id, char * func_compute_weights_key);

/** method signature for weight compuation function used by
    \ref interp_method_callback

  If the field value of a target point, whose weights have been computed
  by this routine, is to be computed on the source process at which this
  routine was called, YAC would use the weights as follows (it is assumed
  that all required source points are available at this process):

  \code{.c}

  // this routine return an index to a source field value associated to
  // the provided global id
  size_t global_id_2_local_idx(int src_field_idx, int global_id);

  [...]

  double tgt_value[collection_size];
  for (int collection_idx = 0; collection_idx < collection_size;
       collection_idx)
    tgt_value[collection_idx] = 0.0;

  // for all source fields
  for (int src_field_idx = 0; src_field_idx < num_pointsets; ++src_field_idx) {

    for (size_t i = 0; i < result_count[src_field_idx]; ++i) {

      size_t src_field_point_idx =
        global_id_2_local_idx(
          src_field_idx, global_results_points[src_field_idx][i]);
      double weight = result_weights[src_field_idx][src_field_point_idx];

      for (int collection_idx = 0; collection_idx < collection_size;
           collection_idx) {

        tgt_value[collection_idx] +=
          src_field[collection_idx][src_field_idx][src_field_point_idx] *
          weight;
      }
    }
  }

  \endcode

  If one or more source points associated to the global ids return by this
  routine are not actually available at this process, YAC will ensure that
  the respective value will be available for the interpolation.

  @param[in]  tgt_coords            3D coordinates of the target point
  @param[in]  src_cell_id           global id of the source cell matching
                                    the target point
  @param[in]  src_cell_idx          (zero based) index of the source
                                    cell matching the target point
  @param[out] global_results_points global ids of source points to be used
                                    for the interpolation of the target point
  @param[out] result_weights        weights to be used for the interpolation
                                    of the target point
  @param[out] result_count          number of source points to be used
                                    for the interpolation of the target point
                                    per pointset
  @param[in]  user_data             user_data pointer provided to
                                    \ref yac_add_compute_weights_callback
  @remark The size of the first dimension of the arrays global_results_points,
          result_weights, and result_count has to be equal to num_pointsets
          provided to \ref yac_cdef_field or \ref yac_cdef_field_mask that was
          used to define the field involved in the coupling that is
          supposed to use this weight computation function.
  @remark 3D coordinates are for a point of a unit sphere
*/
typedef void (*yac_func_compute_weights)(
  double const tgt_coords[3], int src_cell_id, size_t src_cell_idx,
  int const ** global_results_points, double ** result_weights,
  size_t * result_count, void * user_data);

/** registers a callback routine for the computation of weights by the
 *  interpolation method user_callback

     @param[in] compute_weights_callback pointer to a weight computation routine
     @param[in] user_data                data pointer that will be passed to
                                         compute_weights_callback
     @param[in] key                      key for identifing the callback routine
     @remark the callback has to be set on all source processes taking part in
             respective coupling
     @remark the key has to match the one provided in XML configuration file
*/
void yac_add_compute_weights_callback(
  yac_func_compute_weights compute_weights_callback,
  void * user_data, char const * key);

#endif // YAC_INTERFACE_H

